#define G_LOG_DOMAIN "boilerplate-preference"

#include "config.h"

#include "boilerplate-preferences.h"
#include "boilerplate-window.h"
#include "boilerplate-application.h"

#include <gtk/gtk.h>

/**
 * SECTION:boilerplate-preferences
 * @title: BoilerplatePreferences
 * @short_description: A #GtkWindow instance created for #BoilerplateApplication
 * @include: boilerplate-preferences.h
 * @see_also: #GtkWindow
 *
 * The #BoilerplatePreferences is an instance #GtkWindow, and acts as
 * the settings window for the current #GtkApplication.
 **/

typedef struct _BoilerplatePreferencesPrivate
{
    BoilerplateWindow *window;
    GSettings *settings;

    GtkWidget *toolbar;
    GtkWidget *headerbar;
    GtkWidget *content;

    GtkWidget *search;
    GtkWidget *searchbar;
    GtkWidget *searchentry;

    GtkWidget *example_button;
    GtkWidget *example_combo;
} BoilerplatePreferencesPrivate;

struct _BoilerplatePreferences
{
    GtkWindow parent_instance;
};

G_DEFINE_TYPE_WITH_PRIVATE (BoilerplatePreferences, boilerplate_preferences, GTK_TYPE_WINDOW)

static gboolean
_on_key_press (GtkWidget   *widget,
               GdkEventKey *event,
               gpointer     user_data)
{
    BoilerplatePreferencesPrivate *priv = boilerplate_preferences_get_instance_private (user_data);

    switch (event->keyval)
    {
        case GDK_KEY_Escape:
            /* Close preferences window if searchentry does not have focus */
            if (gtk_widget_is_focus (priv->searchentry) == FALSE)
                gtk_widget_destroy (widget);
            else
                gtk_search_bar_set_search_mode (GTK_SEARCH_BAR (priv->searchbar), FALSE);

            return TRUE;
            break;

        default:
            return FALSE;
    }
    return FALSE;
}

static void
search_text_changed (GtkEntry               *entry,
                     BoilerplatePreferences *preferences)
{
    //BoilerplatePreferencesPrivate *priv = boilerplate_preferences_get_instance_private (BOILERPLATE_PREFERENCES (preferences));
    const gchar *text;

    text = gtk_entry_get_text (entry);

    if (text[0] == '\0')
    {
        return;
    }

    //TO_DO: Implement search filter for GtkGrid on priv->content->notebook->page->GtkGrid
}

static void
boilerplate_preferences_finalize (GObject *object)
{
    BoilerplatePreferencesPrivate *priv = boilerplate_preferences_get_instance_private (BOILERPLATE_PREFERENCES (object));

    g_clear_object (&priv->settings);

    G_OBJECT_CLASS (boilerplate_preferences_parent_class)->finalize (object);
}

static void
boilerplate_preferences_class_init (BoilerplatePreferencesClass *klass)
{
    GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (klass);
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = boilerplate_preferences_finalize;

    gtk_widget_class_set_template_from_resource (widget_class,
                                                 PACKAGE_RESOURCES "/ui/preferences.ui");

    gtk_widget_class_bind_template_child_private (widget_class, BoilerplatePreferences, toolbar);
    gtk_widget_class_bind_template_child_private (widget_class, BoilerplatePreferences, headerbar);
    gtk_widget_class_bind_template_child_private (widget_class, BoilerplatePreferences, content);

    gtk_widget_class_bind_template_child_private (widget_class, BoilerplatePreferences, search);
    gtk_widget_class_bind_template_child_private (widget_class, BoilerplatePreferences, searchbar);
    gtk_widget_class_bind_template_child_private (widget_class, BoilerplatePreferences, searchentry);

    gtk_widget_class_bind_template_child_private (widget_class, BoilerplatePreferences, example_button);
    gtk_widget_class_bind_template_child_private (widget_class, BoilerplatePreferences, example_combo);

    gtk_widget_class_bind_template_callback (widget_class, search_text_changed);
}

static void
boilerplate_preferences_init (BoilerplatePreferences *self)
{
    BoilerplatePreferencesPrivate *priv = boilerplate_preferences_get_instance_private (self);

    gtk_widget_init_template (GTK_WIDGET (self));

     /* Set window title */
    gtk_window_set_title (GTK_WINDOW (self), PACKAGE_NAME);

    /* Set default window decorations (a.k.a GtkHeaderBar) */
    gtk_window_set_titlebar (GTK_WINDOW (self), priv->headerbar);
    gtk_container_add (GTK_CONTAINER (priv->headerbar), priv->toolbar);

    /* Connect escape key */
    g_signal_connect (G_OBJECT (self),
                      "key_press_event",
                      G_CALLBACK (_on_key_press),
                      self);

    /* Bind search */
    g_object_bind_property (priv->search, "active",
                            priv->searchbar, "search-mode-enabled",
                            G_BINDING_BIDIRECTIONAL);

    /* Bind settings from GUI */
    priv->settings = g_settings_new (PACKAGE_ID);

    g_settings_bind (priv->settings, "example-button",
                     priv->example_button, "active",
                     G_SETTINGS_BIND_DEFAULT);

    g_settings_bind(priv->settings, "example-combo",
                    priv->example_combo, "active-id",
                    G_SETTINGS_BIND_DEFAULT);
}

/**
 * boilerplate_preferences_new:
 * @window: A #GtkApplicationWindow instance
 *
 * Creates a new #BoilerplatePreferences, a sub-object of #GtkWindow.
 *
 * ## Usage
 * |[
 * BoilerplatePreferences *preferences;
 *
 * preferences = boilerplate_preferences_new (window);
 * ]|
 *
 * When creating a boilerplate_preferences_new() a #BoilerplateWindow will need
 * to be passed. This is because #BoilerplatePreferences acts as a modal window
 * to the master window #BoilerplateWindow attached to the
 * #BoilerplateApplication.
 *
 * Returns: #BoilerplatePreferences; a sub-object of #GtkWindow
 */
BoilerplatePreferences *
boilerplate_preferences_new (BoilerplateWindow *window)
{
    return g_object_new (BOILERPLATE_TYPE_PREFERENCES,
                         "transient-for", window,
                         NULL);
}
